from typing import Dict, List, Union

Agg = Dict[str, str]  # dict of size 1: mapping colomn -> aggregation function


def add_aggregation_columns(df, *, group_cols: Union[str, List[str]], aggregations: Dict[str, Agg]):
    """
    Add new columns containing aggregations values on existing columns

    ---

    ### Parameters

    *mandatory :*
    - `group_cols` (*str* or *list*): columns used to aggregate the data
    - `aggregations` (*dict*): keys are name of new columns and values are aggregation functions
       Examples of aggregation functions : 'sum', 'max'
       Available aggregation functions are listed [here](
       https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html#aggregation)

    ---

    ### Example

    **Input**

    | ENTITY | YEAR | VALUE_1 | VALUE_2 |
    |:------:|:----:|:-------:|:-------:|
    |    A   | 2017 |    10   |    3    |
    |    A   | 2017 |    20   |    1    |
    |    A   | 2018 |    10   |    5    |
    |    A   | 2018 |    30   |    4    |
    |    B   | 2017 |    60   |    4    |
    |    B   | 2017 |    40   |    3    |
    |    B   | 2018 |    50   |    7    |
    |    B   | 2018 |    60   |    6    |

    ```cson
    add_aggregation_columns:
      group_cols: ['ENTITY', 'YEAR']
      aggregations:
        sum_value1:
          VALUE_1: 'sum'  # sum of `VALUE_1` put in `sum_value1` column
        max_value1:
          VALUE_1: 'max'  # max of `VALUE_1` put in `max_value1` column
        mean_value2:
          VALUE_2: 'mean'  # mean of `VALUE_2` put in `mean_value2` column
    ]
    ```

    **Output**

    | ENTITY | YEAR | VALUE_1 | VALUE_2 | sum_value1 | max_value1 | mean_value2 |
    |:------:|:----:|:-------:|:-------:|:----------:|:----------:|:-----------:|
    |    A   | 2017 |    10   |    3    |     30     |     20     |     2.0     |
    |    A   | 2017 |    20   |    1    |     30     |     20     |     2.0     |
    |    A   | 2018 |    10   |    5    |     40     |     30     |     4.5     |
    |    A   | 2018 |    30   |    4    |     40     |     30     |     4.5     |
    |    B   | 2017 |    60   |    4    |    100     |     60     |     3.5     |
    |    B   | 2017 |    40   |    3    |    100     |     60     |     3.5     |
    |    B   | 2018 |    50   |    7    |    110     |     60     |     6.5     |
    |    B   | 2018 |    60   |    6    |    110     |     60     |     6.5     |

    """
    group = df.groupby(group_cols)
    for new_col, aggs in aggregations.items():
        assert len(aggs) == 1
        [(col, agg)] = aggs.items()
        df[new_col] = group[col].transform(agg)
    return df



"""
Implements a series of decorators functions with decorator arguments (logger).
Decorators modify the behaviour of a function to execute code before and after
the original function.

They are a very useful pattern for logging, timing functions, retrying execution,
checking credentials, patching methods or caching for example.


Examples:

    # augment.py
    import logging
    from toucan_data_sdk.utils.decorators import log_message, log_time, log_shapes


    logger = logging.getLogger(__name__)

    @log_shapes(logger)
    @log_message(logger, "Replace affaires with parent affaires")
    @log_time(logger)
    def parse_reseau(df):
        # Do transformations
        return df

Output:
    >>> 2016-06-23 11:21 - INFO - small_apps.demo.preprocess.augment
                                | parse_reseau - time: 50.1 ms
    >>> 2016-06-23 11:21 - INFO - small_apps.demo.preprocess.augment
                                | parse_reseau - Replace affaires with parent affaires
    >>> 2016-06-23 11:21 - INFO - small_apps.demo.preprocess.augment
                                | parse_reseau - [(5821, 210)] -> [(662, 15)]

Note:
    You can apply multiple decorators to the same function.
    The decorators get applied in order from bottom to top.

"""
import collections
import logging
import time
from functools import partial, wraps
from hashlib import md5
from threading import current_thread

import joblib
import pandas as pd

from .helpers import (
    clean_cachedir_old_entries,
    get_func_sourcecode,
    get_orig_function,
    get_param_value_from_func_call,
    resolve_dependencies,
)

_logger = logging.getLogger(__name__)


def catch(logger):
    """
    Decorator to catch an exception and don't raise it.
    Logs information if a decorator failed.

    Note:
        We don't want possible exceptions during logging to be raised.
        This is used to decorate any function that gets executed
        before or after the execution of the decorated function.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception:
                logger.warning(f'Exception raised in decorator: {func.__name__}')

        return wrapper

    return decorator


def _get_dfs(*args, **kwargs):
    values = args + tuple(kwargs.values())
    return [value for value in values if isinstance(value, pd.DataFrame)]


@catch(_logger)
def _get_dfs_shapes(*args, **kwargs):
    return [df.shape for df in _get_dfs(*args, **kwargs)]


@catch(_logger)
def _log_shapes(logger, func_name, input_shapes, output_shapes):
    logger.info(f'{func_name} - {input_shapes} -> {output_shapes}')


@catch(_logger)
def _log_time(logger, func_name, start, end):
    duration = (end - start) * 1000
    logger.info(f'{func_name} - time: {duration:0.1f} ms')


@catch(_logger)
def _log_message(logger, func_name, message):
    logger.info(f'{func_name} - {message}')


def log_message(logger, message=''):
    """
    Decorator to log a message before executing a function
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _log_message(logger, func.__name__, message)
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator


def log_time(logger):
    """
    Decorator to log the execution time of a function
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            _log_time(logger, func.__name__, start, end)
            return result

        return wrapper

    return decorator


def log_shapes(logger):
    """
    Decorator to log the shapes of input and output dataframes

    It considers all the dataframes passed either as arguments or keyword arguments as inputs
    and all the dataframes returned as outputs.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            input_shapes = _get_dfs_shapes(*args, **kwargs)
            result = func(*args, **kwargs)
            output_shapes = _get_dfs_shapes(result)
            _log_shapes(logger, func.__name__, input_shapes, output_shapes)
            return result

        return wrapper

    return decorator


def log(logger=None, start_message='Starting...', end_message='Done...'):
    """
    Basic log decorator
    Can be used as :
    - @log (with default logger)
    - @log(mylogger)
    - @log(start_message='Hello !", logger=mylogger, end_message='Bye !')
    """

    def actual_log(f, real_logger=logger):
        logger = real_logger or _logger

        @wraps(f)
        def timed(*args, **kwargs):
            logger.info(f'{f.__name__} - {start_message}')
            start = time.time()
            res = f(*args, **kwargs)
            end = time.time()
            logger.info(f'{f.__name__} - {end_message} (took {end - start:.2f}s)')
            return res

        return timed

    if callable(logger):
        return actual_log(logger, real_logger=None)
    return actual_log


def domain(domain_name):
    """
    Allow to apply a function f(df: DataFrame) -> DataFrame) on dfs by specifying the key
    E.g instead of writing:
        def process_domain1(dfs):
            df = dfs['domain1']
            # actual process
            dfs['domain1'] = df
            return dfs

    You can write:
        @domain('domain1')
        def process_domain1(df):
            #actual process
            return df
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            dfs, *args = args
            if not isinstance(dfs, dict):
                raise TypeError(f'{dfs} is not a dict')
            df = dfs.pop(domain_name)
            df = func(df, *args, **kwargs)
            return {domain_name: df, **dfs}

        return wrapper

    return decorator


# ~~~ @cache decorator related stuff ~~~


def cache(  # noqa: C901
    requires=None, disabled=False, applied_on_method=False, check_param=True, limit=None
):
    """Avoid to recompute a function if its parameters and its source code doesnt have changed.

    Args:
        requires: list of dependencies (functions or function names)
        disabled (bool): disable the cache mecanism for this function (useful if you
                             only want to use the dependency mecanism)
        applied_on_method (bool): ignore the first argument (useful to ignore "self")
        check_param (True, False or a str): the name of the parameter to check.
                                                False to not check any of them.
                                                True (default) to check all of them.
        limit (int or None): number of cache entries to keep (no limit by default)
    """
    if not requires:
        requires = []
    elif isinstance(requires, collections.Callable):
        requires = [requires]

    if not isinstance(check_param, (bool, str)):
        raise TypeError("'check_param' must be a str (name of the param to check) or a bool")
    if limit is not None and not isinstance(limit, int):
        raise TypeError("'limit' must be an int (number of cache entries to keep) or None")

    # We keep data in the function attributes so that this data
    # is not erased between two calls:
    if not hasattr(cache, 'funcs_references'):
        cache.funcs_references = {}  # dict of {function_name -> function_object (or None)}
    if not hasattr(cache, 'dependencies'):
        cache.dependencies = {}  # dict of {function_name -> [list of function names]}
    if not hasattr(cache, 'memories'):
        cache.memories = {}  # dict of {thread_id -> joblib.Memory object}

    def decorator(func):
        """This code is executed when the augment module is read (when decorator is applied).
        Here we populate cache.funcs_references and cache.dependencies to use them later."""
        cache.funcs_references[func.__name__] = get_orig_function(func)
        dependencies_names = []
        for requirement in requires:
            if isinstance(requirement, collections.Callable):
                req_name = requirement.__name__
                cache.funcs_references[req_name] = get_orig_function(requirement)
            elif requirement not in cache.funcs_references:
                req_name = requirement
                cache.funcs_references[req_name] = None
            dependencies_names.append(req_name)

        cache.dependencies[func.__name__] = dependencies_names

        @wraps(func)
        def wrapper(*args, **kwargs):
            """This code is executed when a decorated function is actually executed.
            It uses the previously built dependency tree (see above)."""
            current_memory = cache.memories.get(current_thread().name)
            if disabled is True or current_memory is None:
                return func(*args, **kwargs)

            # if cache is enabled, we compute the md5 hash of the concatenated source codes
            # of all the dependencies.
            concatenated_source_code = ''
            dependencies = resolve_dependencies(func.__name__, cache.dependencies)
            for func_name in dependencies:
                function = cache.funcs_references[func_name]
                if function is None:
                    raise Exception(f"Can't get source code of function {func_name!r}")
                source_code = get_func_sourcecode(function)
                concatenated_source_code += source_code
            md5_hash = md5(str.encode(concatenated_source_code)).hexdigest()

            # Add extra parameters so that joblib checks they didnt have changed:
            tmp_extra_kwargs = {
                '__func_dependencies_hash__': md5_hash,
                '__original_func_name__': func.__name__,
            }

            if check_param is True:
                kwargs.update(tmp_extra_kwargs)

                if applied_on_method:
                    self_arg, args = args[0], args[1:]

                @wraps(func)
                def f(*args, **kwargs):
                    # delete the extra parameters that the underlying function doesnt expect:
                    for k in tmp_extra_kwargs.keys():
                        del kwargs[k]

                    if applied_on_method:
                        args = (self_arg,) + args
                    return func(*args, **kwargs)

                f = current_memory.cache(f)
                result = f(*args, **kwargs)
            else:
                if isinstance(check_param, str):
                    check_only_param_value = get_param_value_from_func_call(
                        param_name=check_param, func=func, call_args=args, call_kwargs=kwargs
                    )
                    tmp_extra_kwargs['__check_only__'] = check_only_param_value

                @wraps(func)
                def f(**tmp_extra_kwargs):
                    return func(*args, **kwargs)

                f = current_memory.cache(f)
                result = f(**tmp_extra_kwargs)

            if limit is not None:
                clean_cachedir_old_entries(f.store_backend, func.__name__, limit)

            return result

        return wrapper

    return decorator


method_cache = partial(cache, applied_on_method=True)


def setup_cachedir(cachedir, mmap_mode=None, bytes_limit=None):
    """This function injects a joblib.Memory object in the cache() function
    (in a thread-specific slot of its 'memories' attribute)."""
    if not hasattr(cache, 'memories'):
        cache.memories = {}

    memory = joblib.Memory(
        location=cachedir, verbose=0, mmap_mode=mmap_mode, bytes_limit=bytes_limit
    )
    cache.memories[current_thread().name] = memory
    return memory



from toucan_data_sdk.utils.helpers import slugify


def get_category_cols(df, threshold):
    obj_df = df.select_dtypes(include=['object'])
    return [col for col in obj_df.columns if len(obj_df[col].unique()) < threshold]


def get_int_cols(df):
    float_df = df.select_dtypes(include=['floating'])
    return [col for col in float_df.columns if all(x.is_integer() for x in float_df[col])]


def clean_dataframe(df, is_slugify=True, threshold=50, rename_cols=None):
    """
    This method is used to:
    - slugify the column names (if slugify is set to True)
    - convert columns to 'category' (if len(unique) < threshold) or 'int'
    - clean the dataframe and rename if necessary
    """
    if is_slugify:
        df = df.rename(columns=slugify)

    df = df.dropna(axis=1, how='all')
    for column in get_category_cols(df, threshold=threshold):
        df[column] = df[column].astype('category')
    for column in get_int_cols(df):
        df[column] = df[column].astype(int)

    if rename_cols is not None:
        df = df.rename(columns=rename_cols)

    return df



import logging

import pandas as pd


def cumsum(df, new_column: str, column: str, index: list, date_column: str, date_format: str):
    """
    DEPRECATED - please use `compute_cumsum` instead
    """
    logging.getLogger(__name__).warning('DEPRECATED: use compute_cumsum')
    date_temp = '__date_temp__'
    if isinstance(index, str):
        index = [index]
    levels = list(range(0, len(index)))
    df[date_temp] = pd.to_datetime(df[date_column], format=date_format)
    reference_cols = [date_temp, date_column]
    df = df.groupby(index + reference_cols).sum()
    df[new_column] = df.groupby(level=levels)[column].cumsum()
    df.reset_index(inplace=True)
    del df[date_temp]

    return df



from typing import Any, Dict, List, Sequence, Union

import pandas as pd

from toucan_data_sdk.utils.helpers import ParamsValueError, check_params_columns_duplicate


def add_missing_row(
    df: pd.DataFrame,
    id_cols: List[str],
    reference_col: str,
    complete_index: Union[Dict[str, str], Sequence[str]] = None,
    method: str = None,
    cols_to_keep: List[str] = None,
) -> pd.DataFrame:
    """
    Add missing row to a df base on a reference column

    ---

    ### Parameters

    *mandatory :*
    - `id_cols` (*list of str*): names of the columns used to create each group
    - `reference_col` (*str*): name of the column used to identify missing rows

    *optional :*
    - `complete_index` (*list* or *dict*): [A, B, C] a list of values used to add missing rows.
      It can also be a dict to declare a date range.
      By default, use all values of reference_col.
    - `method` (*str*): by default all missing rows are added. The possible values are :
        - `"between"` : add missing rows having their value between min and max values for each group,
        - `"between_and_after"` : add missing rows having their value bigger than min value for each group.
        - `"between_and_before"` : add missing rows having their value smaller than max values for each group.
    - `cols_to_keep` (*list of str*): name of other columns to keep, linked to the reference_col.

    ---

    ### Example

    **Input**

    YEAR | MONTH | NAME
    :---:|:---:|:--:
    2017|1|A
    2017|2|A
    2017|3|A
    2017|1|B
    2017|3|B

    ```cson
    add_missing_row:
      id_cols: ['NAME']
      reference_col: 'MONTH'
    ```

    **Output**

    YEAR | MONTH | NAME
    :---:|:---:|:--:
    2017|1|A
    2017|2|A
    2017|3|A
    2017|1|B
    2017|2|B
    2017|3|B

    """
    if cols_to_keep is None:
        cols_for_index = [reference_col]
    else:
        cols_for_index = [reference_col] + cols_to_keep
    check_params_columns_duplicate(id_cols + cols_for_index)

    if method == 'between' or method == 'between_and_after':
        df['start'] = df.groupby(id_cols)[reference_col].transform(min)
        id_cols += ['start']
    if method == 'between' or method == 'between_and_before':
        df['end'] = df.groupby(id_cols)[reference_col].transform(max)
        id_cols += ['end']

    names = id_cols + cols_for_index
    new_df = df.set_index(names)
    index_values: Union[Any, tuple] = df.groupby(id_cols).sum().index.values

    if complete_index is None:
        complex_index_values: Union[Any, tuple] = df.groupby(cols_for_index).sum().index.values
    elif isinstance(complete_index, dict):
        if complete_index['type'] == 'date':
            freq = complete_index['freq']
            date_format = complete_index['format']
            start = complete_index['start']
            end = complete_index['end']
            if isinstance(freq, dict):
                freq = pd.DateOffset(**{k: int(v) for k, v in freq.items()})
            new_index = pd.date_range(start=start, end=end, freq=freq)
            complex_index_values = new_index.strftime(date_format).values
        else:
            raise ParamsValueError(f'Unknown complete index type: ' f'{complete_index["type"]}')
    else:
        complex_index_values = list(complete_index)

    def get_tuple(x: Union[Any, tuple]) -> tuple:
        if not isinstance(x, tuple):
            return (x,)
        return x

    new_tuples_index: List[tuple] = [
        get_tuple(x) + get_tuple(y) for x in index_values for y in complex_index_values
    ]

    new_index = pd.MultiIndex.from_tuples(new_tuples_index, names=names)
    new_df = new_df.reindex(new_index).reset_index()

    if method == 'between' or method == 'between_and_after':
        new_df = new_df[new_df[reference_col] >= new_df['start']]
        del new_df['start']
    if method == 'between' or method == 'between_and_before':
        new_df = new_df[new_df[reference_col] <= new_df['end']]
        del new_df['end']

    return new_df



from typing import List

import numpy as np
import pandas as pd


def roll_up(
    df,
    levels: List[str],
    groupby_vars: List[str],
    extra_groupby_cols: List[str] = None,
    var_name: str = 'type',
    value_name: str = 'value',
    parent_name: str = 'parent',
    agg_func: str = 'sum',
    drop_levels: List[str] = None,
):
    """
    Creates aggregates following a given hierarchy

    ---

    ### Parameters

    *mandatory :*
    - `levels` (*list of str*): name of the columns composing the hierarchy
      (from the top to the bottom level).
    - `groupby_vars` (*list of str*): name of the columns with value to
      aggregate.
    - `extra_groupby_cols` (*list of str*) optional: other columns used to
      group in each level.

    *optional :*
    - `var_name` (*str*) : name of the result variable column.
      By default, `“type”`.
    - `value_name` (*str*): name of the result value column.
      By default, `“value”`.
    - `parent_name` (*str*): name of the result parent column.
      By default, `"parent"`.
    - `agg_func` (*str*): name of the aggregation operation.
      By default, `“sum”`.
    - `drop_levels` (*list of str*): the names of the levels that you may want
      to discard from the output.
    ---

    ### Example

    **Input**

    |    Region |     City |  Population |
    |:---------:|:--------:|:-----------:|
    |       Idf |     Panam|         200 |
    |       Idf |   Antony |          50 |
    |      Nord |    Lille |          20 |

    ```cson
    roll_up:
      levels: ["Region", "City"]
      groupby_vars: "Population"
    ```

    **Output**

    |    Region |     City |  Population |    value |   type |
    |:---------:|:--------:|:-----------:|:--------:|:------:|
    |       Idf |     Panam|         200 |    Panam |   City |
    |       Idf |   Antony |          50 |   Antony |   City |
    |      Nord |    Lille |          20 |    Lille |   City |
    |       Idf |      Nan |         250 |      Idf | Region |
    |      Nord |      Nan |          20 |     Nord | Region |
    """
    dfs = list()
    groupby_cols_cpy = list(levels)
    levels_cpy = list(levels)
    levels_cpy.reverse()

    extra_groupby_cols = extra_groupby_cols or []
    drop_levels = drop_levels or []
    previous_level = None
    for (idx, top_level) in enumerate(levels_cpy):
        # Aggregation
        gb_df = getattr(
            df.groupby(groupby_cols_cpy + extra_groupby_cols)[groupby_vars], agg_func
        )().reset_index()

        # Melt-like columns
        gb_df[var_name] = top_level
        gb_df[value_name] = gb_df[top_level]
        gb_df[parent_name] = gb_df[levels_cpy[idx + 1]] if idx < len(levels_cpy) - 1 else np.NaN
        dfs.append(gb_df)
        if previous_level in drop_levels:
            del dfs[-2]
        previous_level = top_level

        # Remove one level each time in the groupby: lowest level column needs
        # a groupby with every levels, the next level needs every one except
        # the lowest, etc. until the top level column that needs only itself
        # inside the groupby.
        groupby_cols_cpy.pop()
    return pd.concat(dfs, sort=False).reset_index()



from typing import List, Union


def sort(df, columns: Union[str, List[str]], order: Union[str, List[str]] = 'asc'):
    """

    Sort the data by the value in specified columns

    ---

    ### Parameters

    *mandatory :*
    - `columns` (*str* or *list(str)*): list of columns to order

    *optional :*
    - `order` (*str* or *list(str)*): the ordering condition ('asc' for
    ascending or 'desc' for descending). If not specified, 'asc' by default.
    If a list of columns has been specified for the `columns` parameter,
    the `order` parameter, if explicitly specified, must be a list of same
    length as the `columns` list (if a string is specified, it will be
    replicated in a list of same length of the `columns` list)

    ---

    ### Example

    **Input**

    | variable | value |
    |:--------:|:-----:|
    |     A    |  220  |
    |     B    |  200  |
    |     C    |  300  |
    |     D    |  100  |

    ```cson
    sort:
      columns: 'value'
    ```

    **Output**

    | variable | value |
    |:--------:|:-----:|
    |     D    |  100  |
    |     B    |  200  |
    |     A    |  220  |
    |     C    |  300  |

    """
    if isinstance(columns, str):
        columns = [columns]
    if isinstance(order, str):
        assert order in ['asc', 'desc']
        orders = [order == 'asc'] * len(columns)
    else:
        assert len(order) == len(columns), "'columns' and 'order' lists must be of same length"
        orders = []
        for ord in order:
            assert ord in ['asc', 'desc'], f"Got order value: {order}. Expected 'asc' or 'desc'"
            orders.append(ord == 'asc')
    return df.sort_values(columns, ascending=orders)



from functools import reduce


def predict_number_of_row_from_conf(conf):
    """
    Precict the number of row of the fake data set generated by `fake_data_generator(conf)`
    This is the product of number of values for each label
    """
    num_uniques_by_label = [len(c.get('values', [])) for c in conf if c['type'] == 'label']
    if len(num_uniques_by_label) > 0:
        return reduce(lambda x, y: x * y, num_uniques_by_label)
    else:
        # There are no labels in `conf`. Therefore the generated data will be empty:
        return 0



import logging
import operator as _operator
from typing import List

LOGGER = logging.getLogger(__name__)

MATH_CHARACTERS = '()+-/*%.'
DEPRECATED_COLUMN_QUOTE_CHARS = ('"', "'")
COLUMN_QUOTE_CHARS = ('`',)


def _basic_math_operation(df, new_column, column_1, column_2, op):
    """
    Basic mathematical operation to apply operator on `column_1` and `column_2`
    Both can be either a number or the name of a column of `df`
    Will create a new column named `new_column`
    """
    if not isinstance(column_1, (str, int, float)):
        raise TypeError('column_1 must be a string, an integer or a float')
    if not isinstance(column_2, (str, int, float)):
        raise TypeError('column_2 must be a string, an integer or a float')

    if isinstance(column_1, str):
        column_1 = df[column_1]
    if isinstance(column_2, str):
        column_2 = df[column_2]
    operator = getattr(_operator, op)
    df[new_column] = operator(column_1, column_2)
    return df


def add(df, new_column, column_1, column_2):
    """
    DEPRECATED -  use `formula` instead
    """
    return _basic_math_operation(df, new_column, column_1, column_2, op='add')


def subtract(df, new_column, column_1, column_2):
    """
    DEPRECATED -  use `formula` instead
    """
    return _basic_math_operation(df, new_column, column_1, column_2, op='sub')


def multiply(df, new_column, column_1, column_2):
    """
    DEPRECATED -  use `formula` instead
    """
    return _basic_math_operation(df, new_column, column_1, column_2, op='mul')


def divide(df, new_column, column_1, column_2):
    """
    DEPRECATED -  use `formula` instead
    """
    return _basic_math_operation(df, new_column, column_1, column_2, op='truediv')


def is_float(x):
    try:
        float(x)
    except ValueError:
        return False
    else:
        return True


class Token:
    """
    A formula is a string like this '"2018  " - 2017 + (a - b)'
    In order to parse it, we split it in different tokens and keep track if it was
    quoted or not.
    E.g. in the formula above, `2017` is a number whereas `"2018"` is a column name.
    even though both are strings.
    """

    def __init__(self, text: str, quoted: bool = False):
        self.text = text.strip()
        self.quoted = quoted

    def get_text(self) -> str:
        if not self.quoted and (self.text in MATH_CHARACTERS or is_float(self.text)):
            return self.text
        else:
            return f'`{self.text}`'

    def __len__(self):
        return len(self.text)

    def __repr__(self):
        return repr(self.text)

    def __eq__(self, other):
        return repr(self) == repr(other)


def _parse_formula(formula_str, quote_chars=COLUMN_QUOTE_CHARS) -> List[Token]:
    tokens = []
    current_word = ''
    quote_to_match = None
    for x in formula_str:
        if x in quote_chars and not quote_to_match:
            quote_to_match = x
            continue
        if x == quote_to_match:
            tokens.append(Token(current_word, True))
            current_word = ''
            quote_to_match = None
            continue
        if quote_to_match or x not in MATH_CHARACTERS:
            current_word += x
        else:
            tokens.append(Token(current_word))
            current_word = ''
            tokens.append(Token(x))
    tokens.append(Token(current_word))
    if quote_to_match is not None:
        raise FormulaError('Missing closing quote in formula')
    return [t for t in tokens if t]


def get_new_syntax_formula(formula: str) -> str:
    """
    Now that all columns must explicitly be quoted, we need a function
    to get the new syntax from the deprecated one
    """
    tokens = _parse_formula(formula, quote_chars=DEPRECATED_COLUMN_QUOTE_CHARS)
    return ''.join(t.get_text() for t in tokens)


def formula(df, *, new_column: str, formula: str):
    """
    Do mathematic operations on columns (add, subtract, multiply or divide)

    ---

    ### Parameters

    *mandatory:*
    - `new_column` (*str*): name of the output column
    - `formula` (*str*): Operation on column. Use name of column and special character:
        - `+` for addition
        - `-` for subtraction
        - `*` for multiplication
        - `/` for division

    **Note:**
    your column name must be quoted with backtick `

    ---

    ### Examples

    **Input**

    | variable | valueA | valueB | My rate |
    |:--------:|:--------:|:-----:|:------:|
    |   toto   |    20    |  100  |   10   |
    |   toto   |    30    |  200  |   10   |
    |   toto   |    10    |  300  |   10   |


    ```cson
    formula:
      new_column: 'valueD'
      formula: '(`valueB` + `valueA` ) / `My rate`'
    ```

    **Output**

    | variable | valueA   | valueB |  My rate |  valueD |
    |:--------:|:--------:|:------:|:-------:|:-------:|
    |   toto   |    20    |   100  |    10   |     12  |
    |   toto   |    30    |   200  |    10   |     23  |
    |   toto   |    10    |   300  |    10   |     31  |

    ---

    **Input**

    | variable | 2018 | 2019 |
    |:--------:|:--------:|:-----:|
    |   toto   |    20    |  100  |
    |   toto   |    30    |  200  |
    |   toto   |    10    |  300  |

    ```cson
    formula:
      new_column: 'Evolution'
      formula: "`2019` - `2018`"
    ```

    **Output**

    | variable | 2018 | 2019 | Evolution |
    |:--------:|:--------:|:-----:|:-----:|
    |   toto   |    20    |  100  | 80 |
    |   toto   |    30    |  200  | 170 |
    |   toto   |    10    |  300  | 290 |

    """
    if '`' not in formula:  # OLD SYNTAX
        old_formula = formula
        formula = get_new_syntax_formula(old_formula)
        LOGGER.warning(
            f'DEPRECATED: You should always use ` for your columns. '
            f'Old syntax: {old_formula!r}, new syntax: {formula!r}'
        )
    tokens = _parse_formula(formula)
    expression_splitted = []
    for t in tokens:
        # To use a column name with only digits, it has to be quoted!
        # Otherwise it is considered as a regular number
        if not t.quoted and (t.text in MATH_CHARACTERS or is_float(t.text)):
            expression_splitted.append(t.text)
        elif t.text in df.columns:
            expression_splitted.append(f'df["{t.text}"]')
        else:
            raise FormulaError(f'"{t.text}" is not a valid column name')
    expression = ''.join(expression_splitted)
    df[new_column] = eval(expression)
    return df


class FormulaError(Exception):
    """Raised when a formula is not valid"""


def round_values(df, *, column: str, decimals: int, new_column: str = None):
    """
    Round each value of a column

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): name of the column to round
    - `decimals` (*int*): number of decimal to keeep

    *optional :*
    - `new_column` (*str*): name of the new column to create.
      By default, no new column will be created and `column` will be replaced

    ---

    ### Example

    ** Input**

    ENTITY|VALUE_1|VALUE_2
    :-----:|:-----:|:-----:
    A|-1.512|-1.504
    A|0.432|0.14

    ```cson
    round_values:
      column: 'VALUE_1'
      decimals:1
      new_column: 'Pika'
    ```

    **Output**

    ENTITY|VALUE_1|VALUE_2|Pika
    :-----:|:-----:|:-----:|:-----:
    A|-1.512|-1.504|-1.5
    A|0.432|0.14|0.4
    """
    new_column = new_column or column
    df[new_column] = df[column].round(decimals)
    return df


def absolute_values(df, *, column: str, new_column: str = None):
    """
    Get the absolute numeric value of each element of a column

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): name of the column

    *optional :*
    - `new_column` (*str*): name of the column containing the result.
      By default, no new column will be created and `column` will be replaced.

    ---

    ### Example

    **Input**

    | ENTITY | VALUE_1 | VALUE_2 |
    |:------:|:-------:|:-------:|
    | A      | -1.512  | -1.504  |
    | A      | 0.432   | 0.14    |

    ```cson
    absolute_values:
      column: 'VALUE_1'
      new_column: 'Pika'
    ```

    **Output**

    | ENTITY | VALUE_1 | VALUE_2 | Pika  |
    |:------:|:-------:|:-------:|:-----:|
    | A      | -1.512  | -1.504  | 1.512 |
    | A      | 0.432   | 0.14    | 0.432 |
    """
    new_column = new_column or column
    df[new_column] = abs(df[column])
    return df



from typing import Dict, List, Union


def groupby(
    df, *, group_cols: Union[str, List[str]], aggregations: Dict[str, Union[str, List[str]]]
):
    """
    Aggregate values by groups.

    ---

    ### Parameters

    *mandatory :*
    - `group_cols` (*list*): list of columns used to group data
    - `aggregations` (*dict*): dictionnary of values columns to group as keys and aggregation
      function to use as values (See the [list of aggregation functions](
      https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html#aggregation))

    ---

    ### Example

    **Input**

    | ENTITY | YEAR | VALUE_1 | VALUE_2 |
    |:------:|:----:|:-------:|:-------:|
    |    A   | 2017 |    10   |    3    |
    |    A   | 2017 |    20   |    1    |
    |    A   | 2018 |    10   |    5    |
    |    A   | 2018 |    30   |    4    |
    |    B   | 2017 |    60   |    4    |
    |    B   | 2017 |    40   |    3    |
    |    B   | 2018 |    50   |    7    |
    |    B   | 2018 |    60   |    6    |

    ```cson
    groupby:
      group_cols: ['ENTITY', 'YEAR']
      aggregations:
        'VALUE_1': 'sum',
        'VALUE_2': 'mean'
    ```

    **Output**

    | ENTITY | YEAR | VALUE_1 | VALUE_2 |
    |:------:|:----:|:-------:|:-------:|
    |    A   | 2017 |    30   |   2.0   |
    |    A   | 2018 |    40   |   4.5   |
    |    B   | 2017 |   100   |   3.5   |
    |    B   | 2018 |   110   |   6.5   |

    """
    df = df.groupby(group_cols, as_index=False).agg(aggregations)

    # When several aggregations are performed on the same column, pandas return
    # a multi-indexed dataframe, so we need to flatten the columns index to get
    # back to a unique level header
    if df.columns.nlevels == 2:
        level_0 = df.columns.get_level_values(0)
        level_1 = df.columns.get_level_values(1)
        new_columns = [(f'{x}_{y}' if x else y) for (x, y) in zip(level_1, level_0)]
        df.columns = new_columns
    return df



from typing import List

from toucan_data_sdk.utils.helpers import check_params_columns_duplicate


def compute_ffill_by_group(df, id_cols: List[str], reference_cols: List[str], value_col: str):
    """
    Compute `ffill` with `groupby`
    Dedicated method as there is a performance issue with a simple groupby/fillna (2017/07)
    The method `ffill` propagates last valid value forward to next values.

    ---

    ### Parameters

    *mandatory :*
    - `id_cols` (*list of str*): names of columns used to create each group.
    - `reference_cols` (*list of str*): names of columns used to sort.
    - `value_col` (*str*): name of the columns to fill.

    ---

    ### Example

    **Input**

    name | rank | value
    :------:|:--------------:|:--------:
    A | 1 | 2
    A | 2 | 5
    A | 3 | null
    B | 1 | null
    B | 2 | 7

    ```cson
    compute_ffill_by_group:
      id_cols: ['name']
      reference_cols: ['rank']
      value_col: 'value'
    ```

    **Ouput**

    name | rank | value
    :------:|:--------------:|:--------:
    A | 1 | 2
    A | 2 | 5
    A | 3 | 5
    B | 1 | null
    B | 2 | 7
    """
    check_params_columns_duplicate(id_cols + reference_cols + [value_col])
    df = df.sort_values(by=id_cols + reference_cols)
    df = df.set_index(id_cols)
    df['fill'] = 1 - df[value_col].isnull().astype(int)
    df['fill'] = df.groupby(level=list(range(0, len(id_cols) - 1)))['fill'].cumsum()
    df[value_col] = df[value_col].ffill()
    df.loc[df['fill'] == 0, value_col] = None
    del df['fill']
    return df.reset_index()



from typing import List, Optional
from uuid import uuid4

import pandas as pd

from .filter_by_date import parse_date


def categories_from_dates(
    df,
    date_col: str,
    new_column: str,
    range_steps: List[str],
    category_names: Optional[List[str]] = None,
    date_format: str = '%Y-%m-%d',
):
    """
    Create a new column of categories based on a date column.
    This function will gather into categories dates from the date column
    based on range steps.

    For instance, the dates: [2018-01-02, 2018-01-06, 2018-01-15, 2018-01-16, 2018-01-20]
    with the steps: [2018-01-07, 2018-01-18] will give 3 categories:
    - First category: [2018-01-02, 2018-01-06]
    - Second category: [2018-01-15, 2018-01-16]
    - Third category: [2018-01-20]

    ### Parameters

    *mandatory :*
    - `date_col` (*str*): name of the date column
    - `new_column` (*str*): name of the column for created categories
    - `range_steps` (*list* of *str*): a list of valid dates.
    Each date should be a string matching `date_fmt` and parsable by `strptime`
    but some offset can also be added using `(datestr) + OFFSET` or `(datestr) - OFFSET` syntax.
    When using this syntax, `OFFSET` should be understable by `pandas.Timedelta`
    (see http://pandas.pydata.org/pandas-docs/stable/timedeltas.html)
    and `w`, `week`, `month` and `year` offset keywords are also accepted.
    Additionally, the following symbolic names are supported: `TODAY`, `YESTERDAY`, `TOMORROW`.
    _NOTE_: `datestr` **MUST** be wrapped with parenthesis.

    *optional :*
    - `category_names` (*list* of *str*): names of the categories to be created.
    This list must have a length equal to `range_step` **plus one**
    - `date_format` (*str*): expected date format in column `date_col` (see [available formats](
    https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior)

    ### Example:

    **Input**

    |    date      |    value   |
    |:------------:|:----------:|
    |  2018-01-28  |      1     |
    |  2018-01-28  |      2     |
    |  2018-01-29  |      3     |
    |  2018-01-30  |      1     |
    |  2018-02-01  |      6     |
    |  2019-01-05  |      7     |
    |  2019-01-06  |      7     |
    |  2019-01-07  |      1     |
    |  2019-01-04  |      1     |

    ```cson
    categories_from_dates:
      date_col: 'date'
      new_column: 'categories'
      range_steps: ['(TODAY)-10days', '(TODAY)+1days']
      category_names: ['Old', 'Recent', 'Futur']
    ```

    **Output**

    *Assuming today is 2019-01-06*

    |    date      |    value   | categories |
    |:------------:|:----------:|:----------:|
    |  2018-01-28  |      1     |    Old     |
    |  2018-01-28  |      2     |    Old     |
    |  2018-01-29  |      3     |    Old     |
    |  2018-01-30  |      1     |    Old     |
    |  2018-02-01  |      6     |    Old     |
    |  2019-01-05  |      7     |   Recent   |
    |  2019-01-06  |      7     |   Futur    |
    |  2019-01-07  |      1     |   Futur    |
    |  2019-01-04  |      1     |   Recent   |
    """

    category_names = category_names or [f'Category {i+1}' for i in range(len(range_steps) + 1)]
    if len(range_steps) == 0:
        raise TypeError('range_steps should not have length 0')
    if len(range_steps) + 1 != len(category_names):
        raise TypeError('category_names should have length len(range_steps)+1')

    date_col_at_date_format = str(uuid4())
    df[date_col_at_date_format] = pd.to_datetime(df[date_col], format=date_format)
    df[new_column] = ''

    # first category
    df[new_column][
        df[date_col_at_date_format] < parse_date(range_steps[0], date_format)
    ] = category_names[0]

    # second to before last category
    for i in range(len(range_steps[:-1])):
        start = range_steps[i]
        stop = range_steps[i + 1]
        mask = (df[date_col_at_date_format] < parse_date(stop, date_format)) & (
            df[date_col_at_date_format] >= parse_date(start, date_format)
        )
        df[new_column][mask] = category_names[i + 1]

    # last category
    df[new_column][
        df[date_col_at_date_format] >= parse_date(range_steps[-1], date_format)
    ] = category_names[-1]

    return df.drop(columns=date_col_at_date_format)



# flake8: noqa
from .add_aggregation_columns import add_aggregation_columns
from .argmax import argmax, argmin
from .categories_from_dates import categories_from_dates
from .converter import cast, change_date_format, convert_datetime_to_str, convert_str_to_datetime
from .cumsum import cumsum
from .fillna import fillna
from .filter import drop_duplicates, query
from .filter_by_date import filter_by_date
from .groupby import groupby
from .if_else import if_else
from .math import absolute_values, add, divide, formula, multiply, round_values, subtract
from .melt import melt
from .percentage import percentage
from .pivot import pivot, pivot_by_group
from .rank import rank
from .rename import rename
from .replace import replace
from .sort import sort
from .text import *
from .top import top, top_group
from .waterfall import waterfall



from itertools import product
from typing import List

import numpy as np
import pandas as pd


def fake_data_generator(conf: List[dict]) -> pd.DataFrame:
    """
    `conf` is a list of dictionaries like
      {'type': 'label', 'values': ['Paris', 'Marseille', 'Lyons'], 'name': 'Cities'}
    and each dictionary will add a column.

    There are two different behaviours:
    - type: 'label' -> the new column will be taken into account for a cartesian product
                       with all other labels
    - type: 'number' -> the new column will contain simple numbers
    """

    # First create all the lines with the cartesian product of all the
    # possible values of 'label' columns
    label_confs = [x for x in conf if x['type'] == 'label']
    label_names = [x.get('name', 'Unnamed') for x in label_confs]
    label_values = [x.get('values', []) for x in label_confs]
    df = pd.DataFrame(list(product(*label_values)), columns=label_names)

    # Then add all the 'number' columns
    number_confs = [x for x in conf if x['type'] == 'number']
    for num_conf in number_confs:
        num_column = np.random.uniform(
            low=num_conf.get('min', 0),
            high=num_conf.get('max', num_conf.get('min', 0) + 10),
            size=df.shape[0],
        )
        df[num_conf.get('name', 'Unnamed')] = num_column.round(num_conf.get('digits', 4))

    return df



import inspect
from functools import wraps

import pandas as pd


def _apply_condition(df, condition, new_column):
    """
    `condition` can be a simple string or integer
    but also a dictionary or a list of dictionaries with postprocesses
    """
    import toucan_data_sdk.utils.generic as generic_functions
    import toucan_data_sdk.utils.postprocess as postprocess_functions

    # Single postprocess
    if isinstance(condition, dict):
        condition = [condition]

    # Postprocesses
    if isinstance(condition, list):
        for postprocess_infos in condition:
            postprocess_name = postprocess_infos.pop('postprocess')
            # Retrieve the right postprocess function
            if postprocess_name == 'if_else':
                postprocess_function = if_else
            else:
                postprocess_function = getattr(
                    postprocess_functions, postprocess_name, None
                ) or getattr(generic_functions, postprocess_name)
            # Apply the postprocess to the dataframe
            if 'new_column' not in postprocess_infos:
                postprocess_infos['new_column'] = new_column
            df = postprocess_function(df, **postprocess_infos)
    else:
        # Simple value: condition is a string or a number
        df[new_column] = condition

    return df


def replace_by_reserved_keywords(f):
    """
    As `if`, `else` and `then` are reserved keywords, we have to
    make the mapping to accepted keywords
    """
    reserved_keywords_mapping = {'if_': 'if', 'then_': 'then', 'else_': 'else'}

    @wraps(f)
    def wrapper(*args, **kwargs):
        # Update the params to be sent to `if_else` as `if_`, `else_`, `then_`
        for accepted, reserved in reserved_keywords_mapping.items():
            if reserved in kwargs:
                kwargs[accepted] = kwargs.pop(reserved)
        return f(*args, **kwargs)

    # Update the signature of the wrapper to still have the right documentation
    # and the excepted `if`, `else`, `then` params
    sig_with_accepted_params = inspect.signature(f)
    parameters_with_reserved_names = []

    for param_name, param in sig_with_accepted_params.parameters.items():
        if param_name in reserved_keywords_mapping:
            parameters_with_reserved_names.append(
                param.replace(name=reserved_keywords_mapping[param_name])
            )
        else:
            parameters_with_reserved_names.append(param)

    wrapper.__signature__ = sig_with_accepted_params.replace(
        parameters=parameters_with_reserved_names
    )

    return wrapper


@replace_by_reserved_keywords
def if_else(df, *, if_, then_, else_=None, new_column):
    """
    The usual if...then...else... statement

    ---

    ### Parameters

    *mandatory :*
    - `if` (*str*): string representing the query to filter the data (same as the one used in the postprocess 'query')
      See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html) for more information
    - `then`:
      - *str, int, float*: A row value to use for the new column
      - *dict or list of dicts*: postprocesses functions to have more logic (see examples)
    - `new_column` (*str*): name of the column containing the result.

    *optional :*
    - `else`: same as then but for the non filtered part.
      If not set, the non filtered part won't have any values for the new column

    ---

    ### Example

    ** Input**

    | country  |    city  | clean | the rating |
    |:--------:|:--------:|:-----:|:----------:|
    | France   |  Paris   |   -1  |      3     |
    | Germany  |  Munich  |    4  |      5     |
    | France   |   Nice   |    3  |      4     |
    |  Hell    | HellCity |   -10 |      0     |

    ```cson
    if_else:
      if: '`the rating` % 2 == 0'
      then:
        postprocess: 'formula'
        formula: '(`the rating` + `clean`) / 2'
      else: 'Hey !'
      new_column: 'new'
    ```

    **Output**

    | country  |    city  | clean | the rating | new  |
    |:--------:|:--------:|:-----:|:----------:|:----:|
    | France   |  Paris   |   -1  |      3     | Hey! |
    | Germany  |  Munich  |    4  |      5     | Hey! |
    | France   |   Nice   |    3  |      4     | 3.5  |
    |  Hell    | HellCity |   -10 |      0     | -5.0 |
    """
    # If the index is not unique (e.g. if the dataframe is a concatenation
    # of multiple dataframes), recompute it
    if not df.index.is_unique:
        df.index = pd.RangeIndex(len(df.index))

    if_sub_df = df.query(if_)
    else_sub_df = df[~df.index.isin(if_sub_df.index)]

    if_sub_df = _apply_condition(if_sub_df, then_, new_column)
    else_sub_df = _apply_condition(else_sub_df, else_, new_column)

    new_df = pd.concat([if_sub_df, else_sub_df]).sort_index()

    # Put back the order in columns
    new_cols = [col for col in new_df.columns if col not in df.columns]
    return new_df[[*df.columns, *new_cols]]



from typing import List, Union

import numpy as np


def rank(
    df,
    value_cols: Union[str, List[str]],
    group_cols: List[str] = None,
    rank_cols_names: List[str] = None,
    method='min',
    ascending: bool = True,
):
    """
    This function creates rank columns based on numeric values to be ranked.

    ---

    ### Parameters

    *mandatory :*
    - `value_cols` (*list*): name(s) of the columns used

    *optional :*
    - `group_cols` (*list*): name(s) of the column(s) used to
      create each group inside which independent ranking needs to be applied
    - `rank_cols_names` (*list*): the names of the added ranking columns.
      If not filled, the ranking will be named after the value_cols with a '_rank' suffix
    - `method` (*str*): method to use when encountering equal values:
        - `'min'` (default): lowest rank in group
        - `'max'`: highest rank in group
        - `'average'`: average rank of group
        - `'first'`: ranks assigned in order the values appear in the series
        - `'dense'`: like 'min', but rank always increases by 1 between groups
    - `ascending` (*boolean*): whether the rank should be determined based on
       ascending (default) or descending order

    ---

    ### Example

    **Input**

    | ENTITY | YEAR | VALUE_1 | VALUE_2 |
    | :---: | :---: | :---: | :---: |
    | A | 2017 | 10 | 3 |
    | A | 2017 | 20 | 1 |
    | A | 2018 | 10 | 5 |
    | A | 2018 | 30 | 4 |
    | B | 2017 | 60 | 4 |
    | B | 2017 | 40 | 3 |
    | B | 2018 | 50 | 7 |
    | B | 2018 | 50 | 6 |

    ```cson
    rank :
      value_cols: 'VALUE_1'
    ```

    **Output**

    | ENTITY | YEAR | VALUE_1 | VALUE_2 | VALUE_1_rank
    | :---: | :---: | :---: | :---: | :---: |
    | A | 2017 | 10 | 3 | 1 |
    | A | 2017 | 20 | 1 | 3 |
    | A | 2018 | 10 | 5 | 1 |
    | A | 2018 | 30 | 4 | 4 |
    | B | 2017 | 60 | 4 | 8 |
    | B | 2017 | 40 | 3 | 5 |
    | B | 2018 | 50 | 7 | 6 |
    | B | 2018 | 50 | 6 | 6 |
    """

    value_cols = [value_cols] if not isinstance(value_cols, list) else value_cols
    for col in value_cols:
        if not np.issubdtype(df[col].dtype, np.number):
            raise TypeError(f'{col} specified in value_cols must be of numeric type')

    if rank_cols_names is None:
        rank_cols_names = [x + '_rank' for x in value_cols]

    if group_cols is None:
        df[rank_cols_names] = df[value_cols].rank(method=method, ascending=ascending)
    else:
        df[rank_cols_names] = df.groupby(group_cols)[value_cols].rank(
            method=method, ascending=ascending
        )

    if method != 'average':
        df[rank_cols_names] = df[rank_cols_names].astype('int')

    return df



import pandas as pd


def convert_str_to_datetime(df, *, column: str, format: str):
    """
    Convert string column into datetime column

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): name of the column to format
    - `format` (*str*): current format of the values (see [available formats](
    https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior))
    """
    df[column] = pd.to_datetime(df[column], format=format)
    return df


def convert_datetime_to_str(df, *, column: str, format: str, new_column: str = None):
    """
    Convert datetime column into string column

    ---

    ### Parameters

    *mandatory :*
    - column (*str*): name of the column to format
    - format (*str*): format of the result values (see [available formats](
    https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior))

    *optional :*
    - new_column (*str*): name of the output column. By default `column` is overwritten.
    """
    new_column = new_column or column
    df[new_column] = df[column].dt.strftime(format)
    return df


def change_date_format(
    df,
    *,
    column: str,
    output_format: str,
    input_format: str = None,
    new_column: str = None,
    new_time_zone=None,
):
    """
    Convert the format of a date

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): name of the column to change the format
    - `output_format` (*str*): format of the output values (see [available formats](
    https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior))

    *optional :*
    - `input_format` (*str*): format of the input values (by default let the parser detect it)
    - `new_column` (*str*): name of the output column  (by default overwrite `column`)
    - `new_time_zone` (*str*): name of new time zone (by default no time zone conversion is done)

    ---

    ### Example

    **Input**

    label   | date
    :------:|:----:
    France  | 2017-03-22
    Europe  | 2016-03-22

    ```cson
    change_date_format:
      column: 'date'
      input_format: '%Y-%m-%d'
      output_format: '%Y-%m'
    ```

    Output :

    label   | date
    :------:|:----:
    France  | 2017-03
    Europe  | 2016-03
    """
    new_column = new_column or column
    df[new_column] = (
        pd.to_datetime(df[column], format=input_format, utc=True)
        .dt.tz_convert(new_time_zone)
        .dt.strftime(output_format)
    )
    return df


def cast(df, column: str, type: str, new_column=None):
    """
    Convert column's type into type

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): name of the column to convert
    - `type` (*str*): output type. It can be :
        - `"int"` : integer type
        - `"float"` : general number type
        - `"str"` : text type

    *optional :*
    - `new_column` (*str*): name of the output column.
       By default the `column` arguments is modified.

    ---

    ### Example

    **Input**

    | Column 1 |  Column 2   |  Column 3  |
    |:-------:|:--------:|:--------:|
    |  'one'  |  '2014'  |   30.0   |
    |  'two'  |  2015.0  |    '1'   |
    |   3.1   |   2016   |    450   |

    ```cson
    postprocess: [
      cast:
        column: 'Column 1'
        type: 'str'
      cast:
        column: 'Column 2'
        type: 'int'
      cast:
        column: 'Column 3'
        type: 'float'
    ]
    ```

    **Output**

    | Column 1 |  Column 2  |  Column 3  |
    |:-------:|:------:|:--------:|
    |  'one'  |  2014  |   30.0   |
    |  'two'  |  2015  |    1.0   |
    |  '3.1'  |  2016  |  450.0   |
    """
    new_column = new_column or column
    df[new_column] = df[column].astype(type)
    return df



from typing import Any, List

import numpy as np
import pandas as pd

__all__ = (
    'lower',
    'upper',
    'title',
    'capitalize',
    'swapcase',
    'length',
    'isalnum',
    'isalpha',
    'isdigit',
    'isspace',
    'islower',
    'isupper',
    'istitle',
    'isnumeric',
    'isdecimal',
    'strip',
    'lstrip',
    'rstrip',
    'center',
    'ljust',
    'rjust',
    'split',
    'rsplit',
    'partition',
    'rpartition',
    'find',
    'rfind',
    'index',
    'rindex',
    'startswith',
    'endswith',
    'concat',
    'contains',
    'repeat',
    'replace_pattern',
    # 'slice',
    # 'slice_replace',
    # 'count'
)


###################################################################################################
#                              METHODS WITH NO EXTRA PARAMETERS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :return: the transformed dataframe
###################################################################################################


def _generate_basic_str_postprocess(method_name, docstring):
    def f(df, column: str, new_column: str = None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method()
        return df

    f.__name__ = method_name
    f.__doc__ = f"""
    {docstring}
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.{method_name}.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column

    *optional :*
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    """
    return f


doc = 'Compute length of each string of `column`'
length = _generate_basic_str_postprocess('len', doc)

# lower, upper, capitalize, title, swapcase
###################################################################################################
doc = 'Converts all characters of `column` to lowercase.'
lower = _generate_basic_str_postprocess('lower', doc)

doc = 'Converts all characters of `column` to uppercase.'
upper = _generate_basic_str_postprocess('upper', doc)

doc = (
    'Converts first character to uppercase and remaining ' 'to lowercase for each line of `column`.'
)
capitalize = _generate_basic_str_postprocess('capitalize', doc)

doc = (
    'Converts first character to uppercase and remaining '
    'to lowercase for each word of each line of `column`.'
)
title = _generate_basic_str_postprocess('title', doc)

doc = 'Converts uppercase to lowercase and lowercase to uppercase for each word of `column`.'
swapcase = _generate_basic_str_postprocess('swapcase', doc)

# isalnum, isalpha, isdigit, isspace, islower, isupper, istitle, isnumeric, isdecimal
###################################################################################################
doc = 'Check whether all characters in each string in `column` are alphanumeric'
isalnum = _generate_basic_str_postprocess('isalnum', doc)

doc = 'Check whether all characters in each string in `column` are alphabetic'
isalpha = _generate_basic_str_postprocess('isalpha', doc)

doc = 'Check whether all characters in each string in `column` are digits'
isdigit = _generate_basic_str_postprocess('isdigit', doc)

doc = 'Check whether all characters in each string in `column` are whitespace'
isspace = _generate_basic_str_postprocess('isspace', doc)

doc = 'Check whether all characters in each string in `column` are lowercase'
islower = _generate_basic_str_postprocess('islower', doc)

doc = 'Check whether all characters in each string in `column` are uppercase'
isupper = _generate_basic_str_postprocess('isupper', doc)

doc = 'Check whether all characters in each string in `column` are titlecase'
istitle = _generate_basic_str_postprocess('istitle', doc)

doc = 'Check whether all characters in each string in `column` are numeric'
isnumeric = _generate_basic_str_postprocess('isnumeric', doc)

doc = 'Check whether all characters in each string in `column` are decimal'
isdecimal = _generate_basic_str_postprocess('isdecimal', doc)


###################################################################################################
#                                        STRIP METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param to_strip: (str: None) set of characters to be removed
# :param new_column: the destination column (if not set, `column` will be used)
# :return: the transformed dataframe
###################################################################################################
def _generate_strip_str_postprocess(method_name, docstring):
    def f(df, column: str, *, to_strip: str = None, new_column: str = None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(to_strip)
        return df

    f.__name__ = method_name
    f.__doc__ = f"""
    {docstring}
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.{method_name}.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column

    *optional :*
    - `to_strip` (*str*): set of characters to be removed
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    """
    return f


doc = 'Strip whitespace (including newlines) from each string in `column` from both sides'
strip = _generate_strip_str_postprocess('strip', doc)

doc = 'Strip whitespace (including newlines) from each string in `column` from left side'
lstrip = _generate_strip_str_postprocess('lstrip', doc)

doc = 'Strip whitespace (including newlines) from each string in `column` from left side'
rstrip = _generate_strip_str_postprocess('rstrip', doc)


###################################################################################################
#                              METHODS with `width` and `fillchar`
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param width: (int) minimum width
# :param fillchar: (default: \' \') additional character for filling
# :param new_column: the destination column (if not set, `column` will be used)
# :return: the transformed dataframe
###################################################################################################


def _generate_width_str_postprocess(method_name, docstring):
    def f(df, column: str, *, width: int, fillchar: str = ' ', new_column: str = None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(width, fillchar=fillchar)
        return df

    f.__name__ = method_name
    f.__doc__ = f"""
    {docstring}
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.{method_name}.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column
    - `width` (*int*): minimum widt

    *optional :*
    - `fillchar` (*str*): additional character for filling
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    """
    return f


doc = 'Filling left and right side of strings in `column` with an additional character'
center = _generate_width_str_postprocess('center', doc)

doc = 'Filling right side of strings in `column` with an additional character'
ljust = _generate_width_str_postprocess('ljust', doc)

doc = 'Filling left side of strings in `column` with an additional character'
rjust = _generate_width_str_postprocess('rjust', doc)


###################################################################################################
#                                        SPLIT METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_columns: the destination columns
#        (if not set, columns `column_1`, ..., `column_n` will be created)
# :param sep: (default: \' \') string or regular expression to split on
# :param limit: (default: None) limit number of splits in output
# :return: the transformed dataframe
###################################################################################################
def _generate_split_str_postprocess(method_name, docstring):
    def f(df, column: str, *, new_columns: List[str] = None, sep: str = ' ', limit: int = None):
        method = getattr(df[column].str, method_name)
        df_split = method(pat=sep, n=limit, expand=True)
        nb_cols = df_split.shape[1]
        if new_columns and (not isinstance(new_columns, list) or nb_cols > len(new_columns)):
            raise ValueError(f"'new_columns' should be a list with at least {nb_cols} elements")
        if new_columns is None:
            new_columns = [f'{column}_{i}' for i in range(1, nb_cols + 1)]
        df[new_columns[:nb_cols]] = df_split
        return df

    f.__name__ = method_name
    f.__doc__ = f"""
    {docstring}
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.{method_name}.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column

    *optional :*
    - `sep` (*str*): string or regular expression to split on
    - `limit` (*int*): limit number of splits in output (by default, there is no limit)
    - `new_columns` (*list*): the destination columns (by default, new columns will be added automatically)
    """
    return f


doc = 'Split each string in the caller’s values by given pattern, propagating NaN values'
split = _generate_split_str_postprocess('split', doc)

doc = (
    'Split each string `column` by the given delimiter string, '
    'starting at the end of the string and working to the front'
)
rsplit = _generate_split_str_postprocess('rsplit', doc)


###################################################################################################
#                                        PARTITION METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_columns: the 3 destination columns
# :param sep: (default: \' \') string or regular expression to split on
# :return: the transformed dataframe
###################################################################################################
def _generate_partition_str_postprocess(method_name, docstring):
    def f(df, column: str, *, new_columns: List[str], sep: str = ' '):
        if len(new_columns) != 3:
            raise ValueError('`new_columns` must have 3 columns exactly')
        method = getattr(df[column].str, method_name)
        df[new_columns] = method(sep)
        return df

    f.__name__ = method_name
    f.__doc__ = f"""
    {docstring}
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.{method_name}.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column
    - `new_columns` (*list*): the 3 destination columns

    *optional :*
    - `sep` (*str*): string or regular expression to split on
    """
    return f


doc = (
    'Split the string at the first occurrence of sep, and return 3 elements containing '
    'the part before the separator, the separator itself, and the part after the separator. '
    'If the separator is not found, return 3 elements containing the string itself, '
    'followed by two empty strings.'
)
partition = _generate_partition_str_postprocess('partition', doc)

doc = (
    'Split the string at the last occurrence of sep, and return 3 elements containing '
    'the part before the separator, the separator itself, and the part after the separator. '
    'If the separator is not found, return 3 elements containing two empty strings, '
    'followed by the string itself.'
)
rpartition = _generate_partition_str_postprocess('rpartition', doc)


###################################################################################################
#                                   INDEX AND FIND METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :param sub: substring being searched
# :param start: (default: 0) left edge index
# :param end: (default: None) right edge index
# :return: the transformed dataframe
###################################################################################################
def _generate_find_str_postprocess(method_name, docstring):
    def f(df, column: str, *, sub: str, start: int = 0, end: int = None, new_column: str = None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(sub, start, end)
        return df

    f.__name__ = method_name
    f.__doc__ = f"""
    {docstring}
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.{method_name}.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column
    - `sub` (*str*): substring being searched

    *optional :*
    - `start` (*int*): left edge index
    - `end` (*int*): right edge index
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    """
    return f


doc = (
    'Return lowest indexes in each strings in `column` where the substring '
    'is fully contained between [start:end]. Return -1 on failure'
)
find = _generate_find_str_postprocess('find', doc)

doc = (
    'Return highest indexes in each strings in `column` where the substring '
    'is fully contained between [start:end]. Return -1 on failure'
)
rfind = _generate_find_str_postprocess('rfind', doc)

doc = (
    'Return lowest indexes in each strings where the substring is fully contained '
    'between [start:end]. This is the same as str.find except instead of returning -1, '
    'it raises a ValueError when the substring is not found'
)
index = _generate_find_str_postprocess('index', doc)

doc = (
    'Return highest indexes in each strings where the substring is fully contained '
    'between [start:end]. This is the same as str.find except instead of returning -1, '
    'it raises a ValueError when the substring is not found'
)
rindex = _generate_find_str_postprocess('rindex', doc)


###################################################################################################
#                                  STARTSWITH/ENDSWITH METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :param pat: character sequence
# :param na: (default: NaN) object shown if element tested is not a string
# :return: the transformed dataframe
###################################################################################################
def _generate_with_str_postprocess(method_name, docstring):
    def f(df, column: str, *, pat: str, na: Any = np.nan, new_column: str = None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(pat, na=na)
        return df

    f.__name__ = method_name
    f.__doc__ = (
        f'{docstring}\n'
        f':param df: the dataframe\n'
        f':param column: the column\n'
        f':param new_column: the destination column (if not set, `column` will be used)\n'
        f':param pat: character sequence\n'
        f':param na: (default: NaN) object shown if element tested is not a string\n'
        f':return: the transformed dataframe'
    )
    f.__doc__ = f"""
    {docstring}
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.{method_name}.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column
    - `pat` (*str*): character sequence

    *optional :*
    - `na`: object shown if element tested is not a string
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    """
    return f


doc = 'Test if the start of each string element matches a pattern.'
startswith = _generate_with_str_postprocess('startswith', doc)

doc = 'Test if the end of each string element matches a pattern.'
endswith = _generate_with_str_postprocess('endswith', doc)


###################################################################################################
#                                        OTHER METHODS
###################################################################################################
def concat(df, *, columns: List[str], new_column: str, sep: str = None):
    """
    Concatenate `columns` element-wise
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.cat.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `columns` (*list*): list of columns to concatenate (at least 2 columns)
    - `new_column` (*str*): the destination column

    *optional :*
    - `sep` (*str*): the separator
    """
    if len(columns) < 2:
        raise ValueError('The `columns` parameter needs to have at least 2 columns')
    first_col, *other_cols = columns

    def is_integer_column(df: pd.DataFrame, column: str):
        """Check if a column has only integers or NaN"""
        try:
            return all(pd.isnull(x) or x.is_integer() for x in df[column])
        except Exception:
            return False

    str_sub_df = df[columns]
    for col in columns:
        # In case of integer columns, we don't want the values to be considered
        # as floats (with leading `.0` because of NaN values)
        if is_integer_column(str_sub_df, col):
            str_sub_df[col] = str_sub_df[col].apply(lambda x: '' if pd.isnull(x) else int(x))
        str_sub_df[col] = str_sub_df[col].astype(str)

    df.loc[:, new_column] = str_sub_df[first_col].str.cat(str_sub_df[other_cols], sep=sep)
    return df


def contains(
    df,
    column: str,
    *,
    pat: str,
    new_column: str = None,
    case: bool = True,
    na: Any = None,
    regex: bool = True,
):
    """
    Test if pattern or regex is contained within strings of `column`
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.contains.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column
    - `pat` (*str*): character sequence or regular expression.

    *optional :*
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    - `case` (*boolean*): if true, case sensitive.
    - `na`: fill value for missing values.
    - `regex` (*boolean*): default true
    """
    new_column = new_column or column
    df.loc[:, new_column] = df[column].str.contains(pat, case=case, na=na, regex=regex)
    return df


def repeat(df, column: str, *, times: int, new_column: str = None):
    """
    Duplicate each string in `column` by indicated number of time
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.repeat.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column
    - `times` (*int*): times to repeat the string

    *optional :*
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    """
    new_column = new_column or column
    df.loc[:, new_column] = df[column].str.repeat(times)
    return df


def replace_pattern(
    df,
    column: str,
    *,
    pat: str,
    repl: str,
    new_column: str = None,
    case: bool = True,
    regex: bool = True,
):
    """
    Replace occurrences of pattern/regex in `column` with some other string
    See [pandas doc](
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.replace.html) for more information

    ---

    ### Parameters

    *mandatory :*
    - `column` (*str*): the column
    - `pat` (*str*): character sequence or regular expression
    - `repl` (*str*): replacement string

    *optional :*
    - `new_column` (*str*): the destination column (if not set, `column` will be used)
    - `case` (*boolean*): if true, case sensitive.
    - `regex` (*boolean*): default true
    """
    new_column = new_column or column
    df.loc[:, new_column] = df[column].str.replace(pat, repl, case=case, regex=regex)
    return df
